import abjad


def test_get_leaf_01():
    staff = abjad.Staff(
        [abjad.Voice("c'8 d'8 e'8 f'8"), abjad.Voice("g'8 a'8 b'8 c''8")]
    )

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
            }
            \new Voice
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(staff)
    assert abjad.get.leaf(leaves[0], -1) is None
    assert abjad.get.leaf(leaves[0], 0) is leaves[0]
    assert abjad.get.leaf(leaves[0], 1) is leaves[1]


def test_get_leaf_02():
    """
    Voice.
    """

    voice = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4)])

    assert abjad.get.leaf(voice[0], 1) is voice[1]
    assert abjad.get.leaf(voice[1], 1) is voice[2]
    assert abjad.get.leaf(voice[2], 1) is voice[3]
    assert abjad.get.leaf(voice[3], 1) is None

    assert abjad.get.leaf(voice[0], -1) is None
    assert abjad.get.leaf(voice[1], -1) is voice[0]
    assert abjad.get.leaf(voice[2], -1) is voice[1]
    assert abjad.get.leaf(voice[3], -1) is voice[2]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            cs'8
            d'8
            ef'8
        }
        """
    )


def test_get_leaf_03():
    """
    Staff.
    """

    staff = abjad.Staff([abjad.Note(i, (1, 8)) for i in range(4)])

    assert abjad.get.leaf(staff[0], 1) is staff[1]
    assert abjad.get.leaf(staff[1], 1) is staff[2]
    assert abjad.get.leaf(staff[2], 1) is staff[3]
    assert abjad.get.leaf(staff[3], 1) is None

    assert abjad.get.leaf(staff[0], -1) is None
    assert abjad.get.leaf(staff[1], -1) is staff[0]
    assert abjad.get.leaf(staff[2], -1) is staff[1]
    assert abjad.get.leaf(staff[3], -1) is staff[2]

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            cs'8
            d'8
            ef'8
        }
        """
    )


def test_get_leaf_04():
    """
    Container.
    """

    container = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4)])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            d'8
            ef'8
        }
        """
    )

    assert abjad.get.leaf(container[0], 1) is container[1]
    assert abjad.get.leaf(container[1], 1) is container[2]
    assert abjad.get.leaf(container[2], 1) is container[3]
    assert abjad.get.leaf(container[3], 1) is None

    assert abjad.get.leaf(container[0], -1) is None
    assert abjad.get.leaf(container[1], -1) is container[0]
    assert abjad.get.leaf(container[2], -1) is container[1]
    assert abjad.get.leaf(container[3], -1) is container[2]


def test_get_leaf_05():
    """
    Tuplet.
    """

    tuplet = abjad.Tuplet((2, 3), "c'8 cs'8 d'8")

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tuplet 3/2
        {
            c'8
            cs'8
            d'8
        }
        """
    )

    assert abjad.get.leaf(tuplet[0], 1) is tuplet[1]
    assert abjad.get.leaf(tuplet[1], 1) is tuplet[2]
    assert abjad.get.leaf(tuplet[2], 1) is None

    assert abjad.get.leaf(tuplet[0], -1) is None
    assert abjad.get.leaf(tuplet[1], -1) is tuplet[0]
    assert abjad.get.leaf(tuplet[2], -1) is tuplet[1]


def test_get_leaf_06():
    """
    Contiguous containers inside a voice.
    """

    container_1 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4)])
    container_2 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice = abjad.Voice([container_1, container_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(container_1[0], 1) is container_1[1]
    assert abjad.get.leaf(container_1[1], 1) is container_1[2]
    assert abjad.get.leaf(container_1[2], 1) is container_1[3]
    assert abjad.get.leaf(container_1[3], 1) is container_2[0]

    assert abjad.get.leaf(container_1[1], -1) is container_1[0]
    assert abjad.get.leaf(container_1[2], -1) is container_1[1]
    assert abjad.get.leaf(container_1[3], -1) is container_1[2]
    assert abjad.get.leaf(container_2[0], -1) is container_1[3]


def test_get_leaf_07():
    """
    Tuplets inside a voice.
    """

    tuplet_1 = abjad.Tuplet((2, 3), "c'8 cs'8 d'8")
    tuplet_2 = abjad.Tuplet((2, 3), "ef'8 e'8 f'8")
    voice = abjad.Voice([tuplet_1, tuplet_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tuplet 3/2
            {
                c'8
                cs'8
                d'8
            }
            \tuplet 3/2
            {
                ef'8
                e'8
                f'8
            }
        }
        """
    )

    assert abjad.get.leaf(tuplet_1[0], 1) is tuplet_1[1]
    assert abjad.get.leaf(tuplet_1[1], 1) is tuplet_1[2]
    assert abjad.get.leaf(tuplet_1[2], 1) is tuplet_2[0]

    assert abjad.get.leaf(tuplet_1[1], -1) is tuplet_1[0]
    assert abjad.get.leaf(tuplet_1[2], -1) is tuplet_1[1]
    assert abjad.get.leaf(tuplet_2[0], -1) is tuplet_1[2]


def test_get_leaf_08():
    """
    Does not continue across contiguous anonymous voices inside a staff.
    """

    voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4)])
    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    staff = abjad.Staff([voice_1, voice_2])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            \new Voice
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(voice_1[3], 1) is None
    assert abjad.get.leaf(voice_2[0], -1) is None


def test_get_leaf_09():
    """
    Does cross contiguous equally named voices inside a staff.
    """

    voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4)])
    voice_1.name = "My Voice"
    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice_2.name = "My Voice"
    staff = abjad.Staff([voice_1, voice_2])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "My Voice"
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            \context Voice = "My Voice"
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(voice_1[0], 1) is voice_1[1]
    assert abjad.get.leaf(voice_1[1], 1) is voice_1[2]
    assert abjad.get.leaf(voice_1[2], 1) is voice_1[3]
    assert abjad.get.leaf(voice_1[3], 1) is voice_2[0]

    assert abjad.get.leaf(voice_1[1], -1) is voice_1[0]
    assert abjad.get.leaf(voice_1[2], -1) is voice_1[1]
    assert abjad.get.leaf(voice_1[3], -1) is voice_1[2]
    assert abjad.get.leaf(voice_2[0], -1) is voice_1[3]


def test_get_leaf_10():
    """
    Does not connect through contiguous unequally named voices.
    """

    voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4)])
    voice_1.name = "Your Voice"
    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice_2.name = "My Voice"
    staff = abjad.Staff([voice_1, voice_2])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "Your Voice"
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            \context Voice = "My Voice"
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(voice_1[0], 1) is voice_1[1]
    assert abjad.get.leaf(voice_1[1], 1) is voice_1[2]
    assert abjad.get.leaf(voice_1[2], 1) is voice_1[3]
    assert abjad.get.leaf(voice_1[3], 1) is None

    voice_2.name = None
    assert abjad.get.leaf(voice_1[3], 1) is None

    assert abjad.get.leaf(voice_2[1], -1) is voice_2[0]
    assert abjad.get.leaf(voice_2[2], -1) is voice_2[1]
    assert abjad.get.leaf(voice_2[3], -1) is voice_2[2]
    assert abjad.get.leaf(voice_2[0], -1) is None


def test_get_leaf_11():
    """
    Does connect through like-named staves containing like-named voices.
    """

    voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4)])
    voice_1.name = "low"
    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice_2.name = "low"

    staff_1 = abjad.Staff([voice_1])
    staff_1.name = "mystaff"
    staff_2 = abjad.Staff([voice_2])
    staff_2.name = "mystaff"

    container = abjad.Container([staff_1, staff_2])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \context Staff = "mystaff"
            {
                \context Voice = "low"
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            }
            \context Staff = "mystaff"
            {
                \context Voice = "low"
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
            }
        }
        """
    )

    assert abjad.get.leaf(voice_1[3], 1) is voice_2[0]
    assert abjad.get.leaf(voice_2[0], -1) is voice_1[3]


def test_get_leaf_12():
    """
    Does connect through like-named staves containing like-named voices.
    """

    lower_voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4)])
    lower_voice_1.name = "low"
    lower_voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    lower_voice_2.name = "low"
    higher_voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(12, 16)])
    higher_voice_1.name = "high"
    higher_voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(16, 20)])
    higher_voice_2.name = "high"

    staff_1 = abjad.Staff([higher_voice_1, lower_voice_1])
    staff_1.name = "mystaff"
    staff_1.simultaneous = True
    staff_2 = abjad.Staff([lower_voice_2, higher_voice_2])
    staff_2.name = "mystaff"
    staff_2.simultaneous = True

    container = abjad.Container([staff_1, staff_2])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \context Staff = "mystaff"
            <<
                \context Voice = "high"
                {
                    c''8
                    cs''8
                    d''8
                    ef''8
                }
                \context Voice = "low"
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            >>
            \context Staff = "mystaff"
            <<
                \context Voice = "low"
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
                \context Voice = "high"
                {
                    e''8
                    f''8
                    fs''8
                    g''8
                }
            >>
        }
        """
    )

    assert abjad.get.leaf(lower_voice_1[3], 1) is lower_voice_2[0]
    assert abjad.get.leaf(higher_voice_1[3], 1) is higher_voice_2[0]

    assert abjad.get.leaf(lower_voice_2[0], -1) is lower_voice_1[3]
    assert abjad.get.leaf(higher_voice_2[0], -1) is higher_voice_1[3]


def test_get_leaf_13():
    """
    Does connect through symmetrical nested containers in a voice.
    """

    container_1 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4)])
    container_1 = abjad.Container([container_1])
    container_2 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    container_2 = abjad.Container([container_2])
    voice = abjad.Voice([container_1, container_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            }
            {
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
            }
        }
        """
    )

    assert abjad.get.leaf(container_1[0][0], 1) is container_1[0][1]
    assert abjad.get.leaf(container_1[0][1], 1) is container_1[0][2]
    assert abjad.get.leaf(container_1[0][2], 1) is container_1[0][3]
    assert abjad.get.leaf(container_1[0][3], 1) is container_2[0][0]

    assert abjad.get.leaf(container_2[0][1], -1) is container_2[0][0]
    assert abjad.get.leaf(container_2[0][2], -1) is container_2[0][1]
    assert abjad.get.leaf(container_2[0][3], -1) is container_2[0][2]
    assert abjad.get.leaf(container_2[0][0], -1) is container_1[0][3]


def test_get_leaf_14():
    """
    Tautological parentage asymmetries result in symmetric (balanced) logical
    voice parentage.
    """

    container_1 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4)])
    container_2 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    container_2 = abjad.Container([container_2])
    container_2 = abjad.Container([container_2])
    voice = abjad.Voice([container_1, container_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            {
                {
                    {
                        e'8
                        f'8
                        fs'8
                        g'8
                    }
                }
            }
        }
        """
    )

    assert abjad.get.leaf(container_1[0], 1) is container_1[1]
    assert abjad.get.leaf(container_1[1], 1) is container_1[2]
    assert abjad.get.leaf(container_1[2], 1) is container_1[3]
    assert abjad.get.leaf(container_1[3], 1) is container_2[0][0][0]

    assert abjad.get.leaf(container_2[0][0][1], -1) is container_2[0][0][0]
    assert abjad.get.leaf(container_2[0][0][2], -1) is container_2[0][0][1]
    assert abjad.get.leaf(container_2[0][0][3], -1) is container_2[0][0][2]
    assert abjad.get.leaf(container_2[0][0][0], -1) is container_1[3]


def test_get_leaf_15():
    """
    Tautological parentage asymmetries result in symmetric (balanced) lgoical
    voice parentage.
    """

    container_1 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4)])
    container_1 = abjad.Container([container_1])
    container_1 = abjad.Container([container_1])
    container_2 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice = abjad.Voice([container_1, container_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                {
                    {
                        c'8
                        cs'8
                        d'8
                        ef'8
                    }
                }
            }
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(container_1[0][0][0], 1) is container_1[0][0][1]
    assert abjad.get.leaf(container_1[0][0][1], 1) is container_1[0][0][2]
    assert abjad.get.leaf(container_1[0][0][2], 1) is container_1[0][0][3]
    assert abjad.get.leaf(container_1[0][0][3], 1) is container_2[0]

    assert abjad.get.leaf(container_2[0], -1) is container_1[0][0][3]
    assert abjad.get.leaf(container_2[1], -1) is container_2[0]
    assert abjad.get.leaf(container_2[2], -1) is container_2[1]
    assert abjad.get.leaf(container_2[3], -1) is container_2[2]


def test_get_leaf_16():
    """
    Does connect in sequence of alternating containers and notes.
    """

    container_1 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(2)])
    container_2 = abjad.Container([abjad.Note(i, (1, 8)) for i in range(3, 5)])
    voice = abjad.Voice([container_1, abjad.Note(2, (1, 8)), container_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                cs'8
            }
            d'8
            {
                ef'8
                e'8
            }
        }
        """
    )

    assert abjad.get.leaf(container_1[1], 1) is voice[1]
    assert abjad.get.leaf(voice[1], 1) is container_2[0]

    assert abjad.get.leaf(voice[1], -1) is container_1[1]
    assert abjad.get.leaf(container_2[0], -1) is voice[1]


def test_get_leaf_17():
    """
    Does connect in sequence of alternating tuplets and notes.
    """

    notes = [abjad.Note(i, abjad.Duration(1, 8)) for i in range(3)]
    tuplet_1 = abjad.Tuplet((2, 3), notes)
    notes = [abjad.Note(i, abjad.Duration(1, 8)) for i in range(4, 7)]
    tuplet_2 = abjad.Tuplet((2, 3), notes)
    voice = abjad.Voice([tuplet_1, abjad.Note(3, (1, 8)), tuplet_2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tuplet 3/2
            {
                c'8
                cs'8
                d'8
            }
            ef'8
            \tuplet 3/2
            {
                e'8
                f'8
                fs'8
            }
        }
        """
    )

    assert abjad.get.leaf(tuplet_1[-1], 1) is voice[1]
    assert abjad.get.leaf(voice[1], 1) is tuplet_2[0]

    assert abjad.get.leaf(voice[1], -1) is tuplet_1[-1]
    assert abjad.get.leaf(tuplet_2[0], -1) is voice[1]


def test_get_leaf_18():
    """
    Does connect through asymmetrically nested tuplets.
    """

    inner_tuplet = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    contents = [abjad.Note("c'4"), inner_tuplet, abjad.Note("c'4")]
    tuplet = abjad.Tuplet((2, 3), contents)

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tuplet 3/2
        {
            c'4
            \tuplet 3/2
            {
                c'8
                c'8
                c'8
            }
            c'4
        }
        """
    )

    assert abjad.get.leaf(tuplet[0], 1) is inner_tuplet[0]
    assert abjad.get.leaf(inner_tuplet[-1], 1) is tuplet[-1]
    assert abjad.get.leaf(tuplet[-1], -1) is inner_tuplet[-1]
    assert abjad.get.leaf(inner_tuplet[0], -1) is tuplet[0]


def test_get_leaf_19():
    """
    Returns none in asymmetric logical voice parentage structures.
    """

    voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(3)])
    note = abjad.Note(3, (1, 8))
    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    staff = abjad.Staff([voice_1, note, voice_2])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                cs'8
                d'8
            }
            ef'8
            \new Voice
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(voice_1[-1], 1) is None
    assert abjad.get.leaf(note, 1) is None

    assert abjad.get.leaf(voice_2[0], -1) is None
    assert abjad.get.leaf(note, -1) is None


def test_get_leaf_20():
    """
    Noncontiguous or broken logical voices do not connect.
    """

    voice_1 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(3)])
    voice_1.name = "My Voice"
    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice_2.name = "Your Voice"
    voice_3 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(4, 8)])
    voice_3.name = "My Voice"
    staff = abjad.Staff([voice_1, voice_2, voice_3])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "My Voice"
            {
                c'8
                cs'8
                d'8
            }
            \context Voice = "Your Voice"
            {
                e'8
                f'8
                fs'8
                g'8
            }
            \context Voice = "My Voice"
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    assert abjad.get.leaf(voice_1[-1], 1) is None
    assert abjad.get.leaf(voice_2[-1], 1) is None

    voice_2.name = None

    assert abjad.get.leaf(voice_1[-1], 1) is None
    assert abjad.get.leaf(voice_2[-1], 1) is None

    assert abjad.get.leaf(voice_3[0], -1) is None
    assert abjad.get.leaf(voice_2[0], -1) is None


def test_get_leaf_21():
    """
    Does not connect through nested anonymous voices.
    """

    inner_voice = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(3)])
    outer_voice = abjad.Voice([inner_voice, abjad.Note(3, (1, 8))])

    assert abjad.lilypond(outer_voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \new Voice
            {
                c'8
                cs'8
                d'8
            }
            ef'8
        }
        """
    )

    assert abjad.get.leaf(inner_voice[0], 1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[1], 1) is inner_voice[2]
    assert abjad.get.leaf(inner_voice[2], 1) is None

    assert abjad.get.leaf(inner_voice[1], -1) is inner_voice[0]
    assert abjad.get.leaf(inner_voice[2], -1) is inner_voice[1]
    assert abjad.get.leaf(outer_voice[1], -1) is None


def test_get_leaf_22():
    """
    Does not connect through nested anonymous voices.
    """

    inner_voice = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(1, 4)])
    outer_voice = abjad.Voice([abjad.Note(0, (1, 8)), inner_voice])

    assert abjad.lilypond(outer_voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            \new Voice
            {
                cs'8
                d'8
                ef'8
            }
        }
        """
    )

    assert abjad.get.leaf(inner_voice[0], 1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[1], 1) is inner_voice[2]
    assert abjad.get.leaf(outer_voice[0], 1) is None

    assert abjad.get.leaf(inner_voice[1], -1) is inner_voice[0]
    assert abjad.get.leaf(inner_voice[2], -1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[0], -1) is None


def test_get_leaf_23():
    """
    Does connect through nested equally named voices.
    """

    inner_voice = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(3)])
    inner_voice.name = "My Voice"
    outer_voice = abjad.Voice([inner_voice, abjad.Note(3, (1, 8))])
    outer_voice.name = "My Voice"

    assert abjad.lilypond(outer_voice) == abjad.string.normalize(
        r"""
        \context Voice = "My Voice"
        {
            \context Voice = "My Voice"
            {
                c'8
                cs'8
                d'8
            }
            ef'8
        }
        """
    )

    assert abjad.get.leaf(inner_voice[0], 1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[1], 1) is inner_voice[2]
    assert abjad.get.leaf(inner_voice[2], 1) is outer_voice[1]

    assert abjad.get.leaf(inner_voice[1], -1) is inner_voice[0]
    assert abjad.get.leaf(inner_voice[2], -1) is inner_voice[1]
    assert abjad.get.leaf(outer_voice[1], -1) is inner_voice[-1]


def test_get_leaf_24():
    """
    Does connect through nested equally named voices.
    """

    inner_voice = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(1, 4)])
    inner_voice.name = "My Voice"
    outer_voice = abjad.Voice([abjad.Note(0, (1, 8)), inner_voice])
    outer_voice.name = "My Voice"

    assert abjad.lilypond(outer_voice) == abjad.string.normalize(
        r"""
        \context Voice = "My Voice"
        {
            c'8
            \context Voice = "My Voice"
            {
                cs'8
                d'8
                ef'8
            }
        }
        """
    )

    assert abjad.get.leaf(inner_voice[0], 1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[1], 1) is inner_voice[2]
    assert abjad.get.leaf(outer_voice[0], 1) is inner_voice[0]

    assert abjad.get.leaf(inner_voice[1], -1) is inner_voice[0]
    assert abjad.get.leaf(inner_voice[2], -1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[0], -1) is outer_voice[0]


def test_get_leaf_25():
    """
    Returns none on nested differently named voices.
    """

    inner_voice = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(3)])
    inner_voice.name = "Your Voice"
    outer_voice = abjad.Voice([inner_voice, abjad.Note(3, (1, 8))])
    outer_voice.name = "My Voice"

    assert abjad.lilypond(outer_voice) == abjad.string.normalize(
        r"""
        \context Voice = "My Voice"
        {
            \context Voice = "Your Voice"
            {
                c'8
                cs'8
                d'8
            }
            ef'8
        }
        """
    )

    assert abjad.get.leaf(inner_voice[0], 1) is inner_voice[1]
    assert abjad.get.leaf(inner_voice[1], 1) is inner_voice[2]
    assert abjad.get.leaf(inner_voice[2], 1) is None

    assert abjad.get.leaf(inner_voice[1], -1) is inner_voice[0]
    assert abjad.get.leaf(inner_voice[2], -1) is inner_voice[1]
    assert abjad.get.leaf(outer_voice[1], -1) is None


def test_get_leaf_26():
    """
    Returns none on nested differently named voices.
    """

    voice_2 = abjad.Voice([abjad.Note(i, (1, 8)) for i in range(1, 4)])
    voice_2.name = "Voice 2"
    voice_1 = abjad.Voice([abjad.Note(0, (1, 8)), voice_2])
    voice_1.name = "Voice 1"

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \context Voice = "Voice 1"
        {
            c'8
            \context Voice = "Voice 2"
            {
                cs'8
                d'8
                ef'8
            }
        }
        """
    )

    assert abjad.get.leaf(voice_2[0], 1) is voice_2[1]
    assert abjad.get.leaf(voice_2[1], 1) is voice_2[2]
    assert abjad.get.leaf(voice_1[0], 1) is None

    assert abjad.get.leaf(voice_2[1], -1) is voice_2[0]
    assert abjad.get.leaf(voice_2[2], -1) is voice_2[1]
    assert abjad.get.leaf(voice_1[1], -1) is voice_2[-1]


def test__inspect_are_logical_voice_01():
    """
    Unincorporated leaves do not share a logical voice.
    Unicorporated leaves do not share a root component.
    False if not allow orphans; True if allow orphans.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
    ]
    assert abjad._getlib._are_logical_voice(notes)


def test__inspect_are_logical_voice_02():
    """
    Container and leaves all logical voice.
    """

    container = abjad.Container("c'8 d'8 e'8 f'8")

    r"""
    {
        c'8
        d'8
        e'8
        f'8
    }
    """

    components = abjad.select.components(container)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_03():
    """
    Tuplet and leaves all logical voice.
    """

    tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")

    r"""
    \tuplet 3/2
    {
        c'8
        d'8
        e'8
    }
    """

    components = abjad.select.components(tuplet)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_04():
    """
    Voice and leaves all appear in same logical voice.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")

    r"""
    \new Voice {
        c'8
        d'8
        e'8
        f'8
    }
    """

    components = abjad.select.components(voice)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_05():
    """
    Anonymous staff and leaves all appear in same logical voice.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")

    r"""
    \new Staff {
        c'8
        d'8
        e'8
        f'8
    }
    """

    components = abjad.select.components(staff)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_06():
    """
    Voice, sequential and leaves all appear in same logical voice.
    """

    voice = abjad.Voice(
        r"""
        {
            c'8
            d'8
            e'8
            f'8
        }
        {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
                e'8
                f'8
            }
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    components = abjad.select.components(voice)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_07():
    """
    Anonymous voice, tuplets and leaves all appear in same logical voice.
    """

    voice = abjad.Voice(
        r"""
        \tuplet 3/2
        {
            c'8
            d'8
            e'8
        }
        \tuplet 3/2
        {
            f'8
            g'8
            a'8
        }
        """
    )

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tuplet 3/2
            {
                c'8
                d'8
                e'8
            }
            \tuplet 3/2
            {
                f'8
                g'8
                a'8
            }
        }
        """
    )

    components = abjad.select.components(voice)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_08():
    """
    Logical voice does not extend across anonymous voices.
    """

    staff = abjad.Staff(
        r"""
        \new Voice {
            c'8
            d'8
            e'8
            f'8
        }
        \new Voice {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
            }
            \new Voice
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(staff)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)
    assert not abjad._getlib._are_logical_voice(staff[:])


def test__inspect_are_logical_voice_09():
    """
    Logical voice encompasses across like-named voices.
    """

    staff = abjad.Staff(
        r"""
        \context Voice = "foo" {
            c'8
            d'8
            e'8
            f'8
        }
        \context Voice = "foo" {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "foo"
            {
                c'8
                d'8
                e'8
                f'8
            }
            \context Voice = "foo"
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(staff)
    assert abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_10():
    """
    Logical voice does not extend across differently named voices.
    """

    staff = abjad.Staff(
        r"""
        \context Voice = "foo" {
            c'8
            d'8
        }
        \context Voice = "bar" {
            e'8
            f'8
        }
        """
    )

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \context Voice = "foo"
            {
                c'8
                d'8
            }
            \context Voice = "bar"
            {
                e'8
                f'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(staff)
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_11():
    """
    Logical voice does not across anonymous voices.
    Logical voice does not extend across anonymous staves.
    """

    container = abjad.Container(
        r"""
        \new Staff {
            \new Voice {
                c'8
                d'8
            }
        }
        \new Staff {
            \new Voice {
                e'8
                f'8
            }
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Staff
            {
                \new Voice
                {
                    c'8
                    d'8
                }
            }
            \new Staff
            {
                \new Voice
                {
                    e'8
                    f'8
                }
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_12():
    """
    Logical voice does not extend across anonymous voices.
    Logical voice does not extend across anonymous staves.
    """

    container = abjad.Container(
        r"""
        \new Staff <<
            \new Voice {
                c'8
                d'8
            }
            \new Voice {
                e'8
                f'8
            }
        >>
        \new Staff <<
            \new Voice {
                g'8
                a'8
            }
            \new Voice {
                b'8
                c''8
            }
        >>
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Staff
            <<
                \new Voice
                {
                    c'8
                    d'8
                }
                \new Voice
                {
                    e'8
                    f'8
                }
            >>
            \new Staff
            <<
                \new Voice
                {
                    g'8
                    a'8
                }
                \new Voice
                {
                    b'8
                    c''8
                }
            >>
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert not abjad._getlib._are_logical_voice(leaves[:4])


def test__inspect_are_logical_voice_13():
    """
    Anonymous voice, sequentials and leaves all appear in same logical voice.
    """

    voice = abjad.Voice(
        r"""
        {
            c'8
            d'8
        }
        {
            e'8
            f'8
        }
        """
    )

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
            }
            {
                e'8
                f'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(voice)
    assert abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_14():
    """
    Logical voice can extend across like-named staves.
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        \context Staff = "foo" {
            c'8
            cs'8
            d'8
            ef'8
        }
        \context Staff = "foo" {
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \context Staff = "foo"
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            \context Staff = "foo"
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_15():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        {
            c'8
            d'8
            e'8
            f'8
        }
        \new Voice {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            {
                c'8
                d'8
                e'8
                f'8
            }
            \new Voice
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_16():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        \new Voice {
            c'8
            d'8
            e'8
            f'8
        }
        {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
            }
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_17():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        {
            c'8
            d'8
            e'8
            f'8
        }
        \context Voice = "foo" {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_18():
    """
    Logical voice can not extend acrossdifferently named implicit voices.
    """

    container = abjad.Container(
        r"""
        \context Voice = "foo" {
            c'8
            d'8
            e'8
            f'8
        }
        {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \context Voice = "foo"
            {
                c'8
                d'8
                e'8
                f'8
            }
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_19():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        {
            c'8
            d'8
            e'8
            f'8
        }
        \new Staff {
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            {
                c'8
                d'8
                e'8
                f'8
            }
            \new Staff
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_20():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        \new Staff {
            c'8
            cs'8
            d'8
            ef'8
        }
        {
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Staff
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_21():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        d'8
        ef'8
        \new Voice {
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            d'8
            ef'8
            \new Voice
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_22():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        \new Voice {
            c'8
            cs'8
            d'8
            ef'8
        }
        e'8
        f'8
        fs'8
        g'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Voice
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_23():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        d'8
        ef'8
        \context Voice = "foo" {
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            d'8
            ef'8
            \context Voice = "foo"
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_24():
    """
    Logical voice can not extend across differently named implicit voices.
    NOTE: THIS IS THE LILYPOND LACUNA.
    LilyPond *does* extend logical voice in this case.
    Abjad does not.
    """

    container = abjad.Container(
        r"""
        \context Voice = "foo" {
            c'8
            cs'8
            d'8
            ef'8
        }
        e'8
        f'8
        fs'8
        g'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \context Voice = "foo"
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_25():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        d'8
        ef'8
        \new Voice {
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            d'8
            ef'8
            \new Voice
            {
                e'8
                f'8
                fs'8
                g'8
            }
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_26():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        \new Staff {
            c'8
            d'8
            e'8
            f'8
        }
        g'8
        a'8
        b'8
        c''8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Staff
            {
                c'8
                d'8
                e'8
                f'8
            }
            g'8
            a'8
            b'8
            c''8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_27():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    voice = abjad.Voice([abjad.Note(n, (1, 8)) for n in range(4)])
    container = abjad.Container([voice])
    notes = [abjad.Note(n, (1, 8)) for n in range(4, 8)]
    container = abjad.Container([container] + notes)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            {
                \new Voice
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            }
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_28():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    voice = abjad.Voice([abjad.Note(n, (1, 8)) for n in range(4)])
    voice.name = "foo"
    container = abjad.Container([voice])
    notes = [abjad.Note(n, (1, 8)) for n in range(4, 8)]
    container = abjad.Container([container] + notes)

    r"""
    {
        {
            \context Voice = "foo" {
                c'8
                cs'8
                d'8
                ef'8
            }
        }
        e'8
        f'8
        fs'8
        g'8
    }
    """

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_29():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    voice_1 = abjad.Voice([abjad.Note(n, (1, 8)) for n in range(4)])
    voice_1.name = "foo"
    voice_2 = abjad.Voice([voice_1])
    voice_2.name = "bar"
    notes = [abjad.Note(n, (1, 8)) for n in range(4, 8)]
    container = abjad.Container([voice_2] + notes)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \context Voice = "bar"
            {
                \context Voice = "foo"
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            }
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_30():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    voice_1 = abjad.Voice([abjad.Note(n, (1, 8)) for n in range(4)])
    voice_2 = abjad.Voice([voice_1])
    notes = [abjad.Note(n, (1, 8)) for n in range(4, 8)]
    container = abjad.Container([voice_2] + notes)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Voice
            {
                \new Voice
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            }
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_31():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    notes = [abjad.Note(n, (1, 8)) for n in range(4)]
    voice_1 = abjad.Voice("c''8 c''8 c''8 c''8")
    voice_2 = abjad.Voice("c'8 c'8 c'8 c'8")
    container = abjad.Container([voice_1, voice_2])
    container.simultaneous = True
    container = abjad.Container(notes + [container])

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            d'8
            ef'8
            <<
                \new Voice
                {
                    c''8
                    c''8
                    c''8
                    c''8
                }
                \new Voice
                {
                    c'8
                    c'8
                    c'8
                    c'8
                }
            >>
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert not abjad._getlib._are_logical_voice(leaves[:8])
    assert not abjad._getlib._are_logical_voice(leaves[4:])


def test__inspect_are_logical_voice_32():
    """
    Logical voice can not extend across differently named implicit voices.
    """

    container = abjad.Container(
        r"""
        <<
            \new Voice {
                c'8
                cs'8
                d'8
                ef'8
            }
            \new Voice {
                e'8
                f'8
                fs'8
                g'8
            }
        >>
        af'8
        a'8
        bf'8
        b'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            <<
                \new Voice
                {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
                \new Voice
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
            >>
            af'8
            a'8
            bf'8
            b'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert not abjad._getlib._are_logical_voice(leaves[:8])
    assert not abjad._getlib._are_logical_voice(leaves[4:])


def test__inspect_are_logical_voice_33():
    """
    Logical voice does extend across gaps.
    Logical voice can not extend across differently named voices.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        \new Voice {
            d'8
            ef'8
            \new Voice {
                e'8
                f'8
                fs'8
                g'8
            }
            af'8
            a'8
        }
        bf'8
        b'8
        """
    )

    outer = (0, 1, 10, 11)
    middle = (2, 3, 8, 9)
    inner = (4, 5, 6, 7)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            \new Voice
            {
                d'8
                ef'8
                \new Voice
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
                af'8
                a'8
            }
            bf'8
            b'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice([leaves[i] for i in outer])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in middle])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in inner])
    assert not abjad._getlib._are_logical_voice(leaves[:4])


def test__inspect_are_logical_voice_34():
    """
    Logical voice does extend across gaps.
    Logical voice can not extend across differently named implicit voices.
    """

    staff = abjad.Staff(
        r"""
        c'8
        cs'8
        \new Staff {
            d'8
            ef'8
            \new Staff {
                e'8
                f'8
                fs'8
                g'8
            }
            af'8
            a'8
        }
        bf'8
        b'8
        """
    )

    outer = (0, 1, 10, 11)
    middle = (2, 3, 8, 9)
    inner = (4, 5, 6, 7)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            cs'8
            \new Staff
            {
                d'8
                ef'8
                \new Staff
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
                af'8
                a'8
            }
            bf'8
            b'8
        }
        """
    )

    leaves = abjad.select.leaves(staff)
    assert abjad._getlib._are_logical_voice([leaves[i] for i in outer])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in middle])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in inner])
    assert not abjad._getlib._are_logical_voice(leaves[:4])


def test__inspect_are_logical_voice_35():
    """
    Containers and leaves all appear in same logical voice.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        {
            d'8
            ef'8
            {
                e'8
                f'8
                fs'8
                g'8
            }
            af'8
            a'8
        }
        bf'8
        b'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            {
                d'8
                ef'8
                {
                    e'8
                    f'8
                    fs'8
                    g'8
                }
                af'8
                a'8
            }
            bf'8
            b'8
        }
        """
    )

    components = abjad.select.components(container)
    assert abjad._getlib._are_logical_voice(components)


def test__inspect_are_logical_voice_36():
    """
    Logical voice can not extend across differently named voices.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        {
            {
                \context Voice = "foo" {
                    d'8
                    ef'8
                    e'8
                    f'8
                }
            }
        }
        fs'8
        g'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            {
                {
                    \context Voice = "foo"
                    {
                        d'8
                        ef'8
                        e'8
                        f'8
                    }
                }
            }
            fs'8
            g'8
        }
        """
    )

    outer = (0, 1, 6, 7)
    inner = (2, 3, 4, 5)

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice([leaves[i] for i in outer])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in inner])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_37():
    """
    Logical voice does not extend over differently named voices.
    """

    container = abjad.Container(
        r"""
        {
            {
                \context Voice = "foo" {
                    c'8
                    cs'8
                    d'8
                    ef'8
                }
            }
        }
        e'8
        f'8
        fs'8
        g'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            {
                {
                    \context Voice = "foo"
                    {
                        c'8
                        cs'8
                        d'8
                        ef'8
                    }
                }
            }
            e'8
            f'8
            fs'8
            g'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_38():
    """
    Can not nest across differently named implicit voices.
    """

    container = abjad.Voice(
        r"""
        {
            {
                {
                    c'8
                    cs'8
                    \new Voice {
                        d'8
                        ef'8
                        e'8
                        f'8
                    }
                    fs'8
                    g'8
                }
            }
        }
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                {
                    {
                        c'8
                        cs'8
                        \new Voice
                        {
                            d'8
                            ef'8
                            e'8
                            f'8
                        }
                        fs'8
                        g'8
                    }
                }
            }
        }
        """
    )

    outer = (0, 1, 6, 7)
    inner = (2, 3, 4, 5)

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice([leaves[i] for i in outer])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in inner])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_39():
    """
    Logical voice can not extend across differently named voices.
    """

    voice = abjad.Voice(
        r"""
        c'8
        cs'8
        {
            d'8
            ef'8
            {
                e'8
                f'8
                \context Voice = "bar" {
                    fs'8
                    g'8
                    af'8
                    a'8
                }
                bf'8
                b'8
            }
            c''8
            cs''8
        }
        d''8
        ef''8
        """
    )

    voice.name = "foo"

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \context Voice = "foo"
        {
            c'8
            cs'8
            {
                d'8
                ef'8
                {
                    e'8
                    f'8
                    \context Voice = "bar"
                    {
                        fs'8
                        g'8
                        af'8
                        a'8
                    }
                    bf'8
                    b'8
                }
                c''8
                cs''8
            }
            d''8
            ef''8
        }
        """
    )

    outer = (0, 1, 2, 3, 4, 5, 10, 11, 12, 13, 14, 15)
    inner = (6, 7, 8, 9)

    leaves = abjad.select.leaves(voice)
    assert abjad._getlib._are_logical_voice([leaves[i] for i in outer])
    assert abjad._getlib._are_logical_voice([leaves[i] for i in inner])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_40():
    """
    Logical voice can not extend across differently named anonymous voices.
    """

    container = abjad.Container(
        r"""
        \new Voice {
            c'8
            cs'8
            d'8
            ef'8
        }
        \new Voice {
            e'8
            f'8
            fs'8
            g'8
        }
        af'8
        a'8
        bf'8
        b'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            \new Voice
            {
                c'8
                cs'8
                d'8
                ef'8
            }
            \new Voice
            {
                e'8
                f'8
                fs'8
                g'8
            }
            af'8
            a'8
            bf'8
            b'8
        }
        """
    )

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice(leaves[:4])
    assert abjad._getlib._are_logical_voice(leaves[4:8])
    assert abjad._getlib._are_logical_voice(leaves[8:])
    assert not abjad._getlib._are_logical_voice(leaves[:8])
    assert not abjad._getlib._are_logical_voice(leaves[4:])
    assert not abjad._getlib._are_logical_voice(leaves)


def test__inspect_are_logical_voice_41():
    """
    Logical voice can not extend across differently named anonymous voices.
    """

    container = abjad.Container(
        r"""
        c'8
        cs'8
        <<
            \new Voice {
                d'8
                ef'8
                e'8
                f'8
            }
            \new Voice {
                fs'8
                g'8
                af'8
                a'8
            }
        >>
        bf'8
        b'8
        """
    )

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            cs'8
            <<
                \new Voice
                {
                    d'8
                    ef'8
                    e'8
                    f'8
                }
                \new Voice
                {
                    fs'8
                    g'8
                    af'8
                    a'8
                }
            >>
            bf'8
            b'8
        }
        """
    )

    outer = (0, 1, 10, 11)

    leaves = abjad.select.leaves(container)
    assert abjad._getlib._are_logical_voice([leaves[i] for i in outer])
    assert abjad._getlib._are_logical_voice(leaves[2:6])
    assert abjad._getlib._are_logical_voice(leaves[6:10])
    assert not abjad._getlib._are_logical_voice(leaves[:6])
    assert not abjad._getlib._are_logical_voice(leaves)



import dataclasses
import typing

from . import tag as _tag


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tweak:
    """
    Tweak.
    """

    string: str
    i: int | None = None
    tag: _tag.Tag | None = None

    def __post_init__(self):
        assert isinstance(self.string, str), repr(self.string)
        if self.i is not None:
            assert isinstance(self.i, int), repr(self.i)
        if self.tag is not None:
            assert isinstance(self.tag, _tag.Tag), repr(self.tag)
        self._parse()

    def _list_contributions(self):
        result = []
        deactivate = False
        strings = [self.string]
        if self.tag is not None:
            strings = _tag.double_tag(strings, self.tag, deactivate=deactivate)
        result.extend(strings)
        return result

    def _parse(self):
        parts = self.string.split()
        post_event = False
        if parts[0] == "-":
            post_event = True
            parts.pop(0)
        assert parts[0] == r"\tweak", repr(self.string)
        parts.pop(0)
        attribute = parts[0]
        parts.pop(0)
        value = " ".join(parts)
        return post_event, attribute, value

    def attribute(self) -> str:
        post_event, attribute, value = self._parse()
        return attribute

    def post_event(self) -> bool:
        post_event, attribute, value = self._parse()
        return post_event

    def value(self) -> str:
        post_event, attribute, value = self._parse()
        return value


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Bundle:
    r"""
    Bundled indicator.

    ..  container:: example

        Raises exception on duplicate attributes:

        >>> abjad.Bundle(
        ...     indicator=abjad.Articulation("."),
        ...     tweaks=(
        ...         abjad.Tweak(r"- \tweak color #blue"),
        ...         abjad.Tweak(r"- \tweak color #red"),
        ...     ),
        ... )
        Traceback (most recent call last):
            ...
        Exception: duplicate 'color' attribute.

    """

    indicator: typing.Any
    tweaks: tuple[Tweak, ...] = ()
    comment: str | None = None

    def __post_init__(self):
        assert not isinstance(self.indicator, Bundle), repr(self.indicator)
        assert isinstance(self.tweaks, tuple), repr(self.tweaks)
        assert all(isinstance(_, Tweak) for _ in self.tweaks)
        attributes = [_.attribute() for _ in self.tweaks]
        for attribute in attributes:
            if 1 < attributes.count(attribute):
                raise Exception(f"duplicate {attribute!r} attribute.")
        assert isinstance(self.comment, str | None), repr(self.comment)

    def _get_contributions(self, *, component=None, wrapper=None):
        try:
            contributions = self.indicator._get_contributions(
                component=component, wrapper=wrapper
            )
        except TypeError:
            component = component or wrapper.component
            contributions = self.indicator._get_contributions(component=component)
        lists = contributions.get_contribution_lists()
        if len(lists) == 2 and ["<>"] in lists:
            lists.remove(["<>"])
        if len(lists) == 2 and [r"\pitchedTrill"] in lists:
            lists.remove([r"\pitchedTrill"])
        assert len(lists) == 1, repr(lists)
        list_ = lists[0]
        strings = []
        if self.comment is not None:
            strings.append(self.comment)
        for tweak in sorted(self.tweaks):
            strings.extend(tweak._list_contributions())
        if 2 <= len(list_) and list_[-2] in ("^", "_", "-"):
            list_[-2:-2] = strings
        else:
            list_[-1:-1] = strings
        return contributions

    def get_attribute(self, attribute: str) -> Tweak | None:
        r"""
        Gets tweak with ``attribute``.

        ..  container:: example

            >>> markup = abjad.Markup(r"\markup Allegro")
            >>> bundle = abjad.bundle(
            ...     markup,
            ...     r"- \tweak color #red",
            ...     r"- \tweak font-size 3",
            ... )
            >>> bundle.get_attribute("color")
            Tweak(string='- \\tweak color #red', i=None, tag=None)

            >>> bundle.get_attribute("style") is None
            True

        """
        tweaks = [_ for _ in self.tweaks if _.attribute() == attribute]
        assert len(tweaks) in (0, 1)
        if tweaks:
            tweak = tweaks[0]
            return tweak
        return None

    def remove(self, tweak: Tweak) -> "Bundle":
        r"""
        Removes ``tweak`` from bundle and returns new bundle.

        ..  container:: example

            >>> markup = abjad.Markup(r"\markup Allegro")
            >>> bundle_1 = abjad.bundle(markup, r"- \tweak color #red")
            >>> tweak = bundle_1.get_attribute("color")

            >>> bundle_2 = bundle_1.remove(tweak)
            >>> bundle_2
            Bundle(indicator=Markup(string='\\markup Allegro'), tweaks=(), comment=None)

            >>> bundle_3 = abjad.bundle(bundle_2, r"- \tweak color #blue")
            >>> bundle_3.tweaks
            (Tweak(string='- \\tweak color #blue', i=None, tag=None),)

        """
        assert tweak in self.tweaks, repr(tweak)
        tweaks = list(self.tweaks)
        tweaks.remove(tweak)
        new_bundle = dataclasses.replace(self, tweaks=tuple(tweaks))
        return new_bundle


def bundle(
    indicator: typing.Any,
    *tweaks: str | Tweak,
    comment: str | None = None,
    overwrite: bool = False,
    tag: _tag.Tag | None = None,
) -> Bundle:
    r"""
    Bundles ``indicator`` with ``tweaks``.

    ..  container:: example

        Bundles indicator:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> bundle = abjad.bundle(
        ...     abjad.Articulation("."),
        ...     r"- \tweak color #red",
        ... )
        >>> abjad.attach(bundle, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                - \staccato
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Bundles existing bundle:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> bundle = abjad.bundle(abjad.Articulation("."), r"- \tweak color #red")
        >>> bundle = abjad.bundle(bundle, r"- \tweak font-size 3")
        >>> abjad.attach(bundle, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                - \tweak font-size 3
                - \staccato
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Raises exception on duplicate attribute:

        >>> bundle = abjad.bundle(
        ...     abjad.Articulation("."),
        ...     r"- \tweak color #blue",
        ...     r"- \tweak color #red",
        ... )
        Traceback (most recent call last):
            ...
        Exception: duplicate 'color' attribute:
        Tweak(string='- \\tweak color #blue', i=None, tag=None)
        Tweak(string='- \\tweak color #red', i=None, tag=None)

        Unless ``overwrite=True``:

        >>> bundle = abjad.bundle(
        ...     abjad.Articulation("."),
        ...     r"- \tweak color #blue",
        ...     r"- \tweak color #red",
        ...     overwrite=True,
        ... )
        >>> for _ in bundle.tweaks: _
        Tweak(string='- \\tweak color #red', i=None, tag=None)

    ..  container:: example

        Also raises exception on duplicate attribute:

        >>> bundle = abjad.bundle(
        ...     abjad.Articulation("."),
        ...     r"- \tweak color #blue",
        ... )
        >>> bundle = abjad.bundle(
        ...     bundle,
        ...     r"- \tweak color #red",
        ... )
        Traceback (most recent call last):
            ...
        Exception: duplicate 'color' attribute:
        OLD: Tweak(string='- \\tweak color #blue', i=None, tag=None)
        NEW: Tweak(string='- \\tweak color #red', i=None, tag=None)

        Unless ``overwrite=True``:

        >>> bundle = abjad.bundle(
        ...     abjad.Articulation("."),
        ...     r"- \tweak color #blue",
        ... )
        >>> bundle = abjad.bundle(
        ...     bundle,
        ...     r"- \tweak color #red",
        ...     overwrite=True,
        ... )
        >>> for _ in bundle.tweaks: _
        Tweak(string='- \\tweak color #red', i=None, tag=None)

    """
    input_tweaks: list[Tweak] = []
    for item in tweaks:
        if isinstance(item, Tweak):
            tweak = item
        else:
            assert isinstance(item, str)
            tweak = Tweak(item, tag=tag)
        tweak_attribute = tweak.attribute()
        for input_tweak in input_tweaks[:]:
            if input_tweak.attribute() == tweak_attribute:
                if overwrite is True:
                    input_tweaks.remove(input_tweak)
                else:
                    message = f"duplicate {tweak_attribute!r} attribute:\n"
                    message += repr(input_tweak) + "\n"
                    message += repr(tweak)
                    raise Exception(message)
        input_tweaks.append(tweak)
    if isinstance(indicator, Bundle):
        bundle_tweaks = list(indicator.tweaks)
        for input_tweak in input_tweaks:
            input_tweak_attribute = input_tweak.attribute()
            for bundle_tweak in bundle_tweaks[:]:
                if bundle_tweak.attribute() == input_tweak_attribute:
                    if overwrite is True:
                        bundle_tweaks.remove(bundle_tweak)
                    else:
                        message = f"duplicate {input_tweak.attribute()!r} attribute:\n"
                        message += f"OLD: {bundle_tweak!r}\n"
                        message += f"NEW: {input_tweak!r}"
                        raise Exception(message)
            bundle_tweaks.append(input_tweak)
        indicator = indicator.indicator
        input_tweaks = bundle_tweaks
    input_tweaks.sort()
    return Bundle(indicator, tweaks=tuple(input_tweaks), comment=comment)


def tweak(
    indicator: typing.Any,
    *tweaks: str | Tweak,
    overwrite: bool = False,
    tag: _tag.Tag | None = None,
) -> None:
    """
    Appends ``tweaks`` to ``indicator``.
    """
    from . import score as _score

    prototype = (
        _score.NoteHead,
        _score.Tuplet,
    )
    assert isinstance(indicator, prototype), repr(indicator)
    try:
        tweaks_ = list(indicator.tweaks)
    except AttributeError:
        raise Exception(indicator)
    if tag is not None:
        assert isinstance(tag, _tag.Tag), repr(tag)
    for item in tweaks:
        duplicate = False
        if isinstance(item, Tweak):
            tweak = item
            if tag is not None:
                tweak = Tweak(tweak.string, tag=tag)
        else:
            assert isinstance(item, str), repr(item)
            if getattr(indicator, "post_event", False) and not item.startswith("-"):
                name = type(indicator).__name__
                message = (
                    f"Must prefix {name} (LilyPond 'post-event') tweak with hyphen."
                )
                raise Exception(message)
            if item.startswith("-") and not getattr(indicator, "post_event", False):
                name = type(indicator).__name__
                message = f"Must not prefix {name} tweak with hyphen."
                raise Exception(message)
            tweak = Tweak(item, tag=tag)
        for existing_tweak in tweaks_[:]:
            if existing_tweak == tweak:
                duplicate = True
                continue
            if existing_tweak.attribute() == tweak.attribute():
                if overwrite is True:
                    tweaks_.remove(existing_tweak)
                else:
                    message = "conflicting tweaks:\n"
                    message += f"    {existing_tweak.string}\n"
                    message += f"    {tweak.string}"
                    raise Exception(message)
        if not duplicate:
            tweaks_.append(tweak)
    tweaks_.sort()
    try:
        indicator.tweaks = tuple(tweaks_)
    except dataclasses.FrozenInstanceError:
        raise Exception(indicator, tweaks_)



import collections
import enum
import typing

from . import _getlib
from . import indicators as _indicators
from . import score as _score
from . import select as _select


def _coerce_exclude(exclude):
    if exclude is None:
        exclude = ()
    elif isinstance(exclude, str | enum.Enum):
        exclude = (exclude,)
    else:
        exclude = tuple(exclude)
    assert isinstance(exclude, tuple), repr(exclude)
    return exclude


def _get_leaf(argument, n: int = 0):
    if n not in (-1, 0, 1):
        message = "n must be -1, 0 or 1:\n"
        message += f"   {repr(n)}"
        raise Exception(message)
    if isinstance(argument, _score.Leaf):
        candidate = _get_sibling_with_graces(argument, n)
        if isinstance(candidate, _score.Leaf):
            return candidate
        return _getlib._get_leaf_from_leaf(argument, n)
    if 0 <= n:
        reverse = False
    else:
        reverse = True
        n = abs(n) - 1
    leaves = _public_iterate_leaves(argument, reverse=reverse)
    for i, leaf in enumerate(leaves):
        if i == n:
            return leaf
    return None


def _get_logical_tie_leaves(leaf):
    if _is_unpitched(leaf):
        return [leaf]
    leaves_before, leaves_after = [], []
    current_leaf = leaf
    while True:
        previous_leaf = _get_leaf(current_leaf, -1)
        if previous_leaf is None:
            break
        if _is_unpitched(previous_leaf):
            break
        if current_leaf._has_indicator(
            _indicators.RepeatTie
        ) or previous_leaf._has_indicator(_indicators.Tie):
            leaves_before.insert(0, previous_leaf)
        else:
            break
        current_leaf = previous_leaf
    current_leaf = leaf
    while True:
        next_leaf = _get_leaf(current_leaf, 1)
        if next_leaf is None:
            break
        if _is_unpitched(next_leaf):
            break
        if current_leaf._has_indicator(_indicators.Tie) or next_leaf._has_indicator(
            _indicators.RepeatTie
        ):
            leaves_after.append(next_leaf)
        else:
            break
        current_leaf = next_leaf
    leaves = leaves_before + [leaf] + leaves_after
    return leaves


def _is_obgc_nongrace_voice(argument):
    if (
        type(argument).__name__ == "Voice"
        and type(argument._parent).__name__ == "Container"
        and type(argument._parent[0]).__name__ == "OnBeatGraceContainer"
    ):
        return True
    return False


def _get_sibling_with_graces(component, n):
    assert n in (-1, 0, 1), repr(component, n)
    if n == 0:
        return component
    if component._parent is None:
        return None
    if component._parent.simultaneous:
        return None
    if (
        n == 1
        and getattr(component._parent, "_main_leaf", None)
        and component._parent._main_leaf._before_grace_container is component._parent
        and component is component._parent[-1]
    ):
        return component._parent._main_leaf
    # component is final leaf in obgc
    if (
        n == 1
        and component is component._parent[-1]
        and component._parent.__class__.__name__ == "OnBeatGraceContainer"
    ):
        return component._parent.get_first_nongrace_leaf()
    if (
        n == 1
        and getattr(component._parent, "_main_leaf", None)
        and component._parent._main_leaf._after_grace_container is component._parent
        and component is component._parent[-1]
    ):
        main_leaf = component._parent._main_leaf
        if main_leaf is main_leaf._parent[-1]:
            return None
        index = main_leaf._parent.index(main_leaf)
        return main_leaf._parent[index + 1]
    if n == 1 and getattr(component, "_after_grace_container", None):
        return component._after_grace_container[0]
    if (
        n == -1
        and getattr(component._parent, "_main_leaf", None)
        and component._parent._main_leaf._after_grace_container is component._parent
        and component is component._parent[0]
    ):
        return component._parent._main_leaf
    if (
        n == -1
        and getattr(component._parent, "_main_leaf", None)
        and component._parent._main_leaf._before_grace_container is component._parent
        and component is component._parent[0]
    ):
        main_leaf = component._parent._main_leaf
        if main_leaf is main_leaf._parent[0]:
            return None
        index = main_leaf._parent.index(main_leaf)
        return main_leaf._parent[index - 1]
    # component is first leaf in obgc nongrace voice
    if (
        n == -1
        and component is component._parent[0]
        and type(component._parent).__name__ == "Voice"
        and type(component._parent._parent).__name__ == "Container"
        and type(component._parent._parent[0]).__name__ == "OnBeatGraceContainer"
    ):
        obgc = component._parent._parent[0]
        return obgc[-1]
    if n == -1 and getattr(component, "_before_grace_container", None):
        return component._before_grace_container[-1]
    index = component._parent.index(component) + n
    if not (0 <= index < len(component._parent)):
        return None
    candidate = component._parent[index]
    if n == 1 and getattr(candidate, "_before_grace_container", None):
        return candidate._before_grace_container[0]
    if n == 1 and candidate.__class__.__name__ == "IndependentAfterGraceContainer":
        return candidate[0]
    if n == -1 and getattr(candidate, "_after_grace_container", None):
        return candidate._after_grace_container[-1]
    if n == -1 and candidate.__class__.__name__ == "IndependentAfterGraceContainer":
        return candidate[-1]
    return candidate


def _is_unpitched(leaf):
    if hasattr(leaf, "written_pitch"):
        return False
    if hasattr(leaf, "written_pitches"):
        return False
    return True


def _iterate_components(
    argument,
    prototype=None,
    *,
    exclude=None,
    do_not_iterate_grace_containers=None,
    grace=None,
    reverse=None,
):
    prototype = prototype or _score.Component
    before_grace_container = None
    after_grace_container = None
    exclude = _coerce_exclude(exclude)
    assert isinstance(exclude, tuple), repr(exclude)
    if grace is not False and isinstance(argument, _score.Leaf):
        before_grace_container = argument._before_grace_container
        after_grace_container = argument._after_grace_container
    if not reverse:
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and before_grace_container
        ):
            yield from _iterate_components(
                before_grace_container,
                prototype,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, prototype):
            if (
                grace is None
                or (grace is True and _getlib._get_grace_container(argument))
                or (grace is False and not _getlib._get_grace_container(argument))
            ):
                if not _should_exclude(argument, exclude):
                    yield argument
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and after_grace_container
        ):
            yield from _iterate_components(
                after_grace_container,
                prototype,
                exclude=exclude,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, collections.abc.Iterable):
            for item in argument:
                yield from _iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                    grace=grace,
                    reverse=reverse,
                )
    else:
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and after_grace_container
        ):
            yield from _iterate_components(
                after_grace_container,
                prototype,
                exclude=exclude,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, prototype):
            if (
                grace is None
                or (grace is True and _getlib._get_grace_container(argument))
                or (grace is False and not _getlib._get_grace_container(argument))
            ):
                if not _should_exclude(argument, exclude):
                    yield argument
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and before_grace_container
        ):
            yield from _iterate_components(
                before_grace_container,
                prototype,
                exclude=exclude,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, collections.abc.Iterable):
            for item in reversed(argument):
                yield from _iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                    grace=grace,
                    reverse=reverse,
                )


def _iterate_descendants(component, cross_offset=None):
    assert isinstance(component, _score.Component), repr(component)
    if component is None:
        components = ()
    else:
        generator = _public_iterate_components(component)
        components = list(generator)
    result = []
    if cross_offset is None:
        result = components
    else:
        for component in components:
            append_x = True
            if not (
                component._get_timespan().start_offset < cross_offset
                and cross_offset < component._get_timespan().stop_offset
            ):
                append_x = False
            if append_x:
                result.append(component)
    return tuple(result)


def _iterate_logical_ties(
    argument,
    *,
    exclude=None,
    grace=None,
    nontrivial=None,
    pitched=None,
    reverse=None,
) -> typing.Iterator[_select.LogicalTie]:
    yielded_logical_ties = set()
    for leaf in _public_iterate_leaves(
        argument, exclude=exclude, grace=grace, pitched=pitched, reverse=reverse
    ):
        leaves = _get_logical_tie_leaves(leaf)
        if leaf is not leaves[0]:
            continue
        if (
            nontrivial is None
            or (nontrivial is True and not len(leaves) == 1)
            or (nontrivial is False and len(leaves) == 1)
        ):
            logical_tie = _select.LogicalTie(leaves)
            if logical_tie not in yielded_logical_ties:
                yielded_logical_ties.add(logical_tie)
                yield logical_tie


def _public_iterate_components(
    argument, prototype=None, *, exclude=None, grace=None, reverse=None
):
    if isinstance(argument, _score.Container):
        for component in _iterate_components(
            argument,
            prototype,
            exclude=exclude,
            do_not_iterate_grace_containers=False,
            grace=grace,
            reverse=reverse,
        ):
            yield component
    elif isinstance(argument, collections.abc.Iterable):
        if not reverse:
            for item in argument:
                generator = _public_iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    grace=grace,
                    reverse=reverse,
                )
                yield from generator
        else:
            for item in reversed(argument):
                generator = _public_iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    grace=grace,
                    reverse=reverse,
                )
                yield from generator
    else:
        for component in _iterate_components(
            argument,
            prototype,
            exclude=exclude,
            do_not_iterate_grace_containers=True,
            grace=grace,
            reverse=reverse,
        ):
            yield component


def _public_iterate_leaves(
    argument,
    prototype=None,
    *,
    exclude=None,
    grace=None,
    pitched=None,
    reverse=None,
):
    prototype = prototype or _score.Leaf
    if pitched is True:
        prototype = (_score.Chord, _score.Note)
    elif pitched is False:
        prototype = (_score.MultimeasureRest, _score.Rest, _score.Skip)
    return _public_iterate_components(
        argument, prototype=prototype, exclude=exclude, grace=grace, reverse=reverse
    )


def _should_exclude(argument, exclude):
    assert isinstance(exclude, tuple)
    for string in exclude:
        if argument._has_indicator(string):
            return True
    return False



import copy

import abjad


def test_Skip___copy___01():
    """
    Copy skip.
    """

    skip_1 = abjad.Skip((1, 4))
    skip_2 = copy.copy(skip_1)

    assert isinstance(skip_1, abjad.Skip)
    assert isinstance(skip_2, abjad.Skip)
    assert abjad.lilypond(skip_1) == abjad.lilypond(skip_2)
    assert skip_1 is not skip_2


def test_Skip___copy___02():
    """
    Copy skip with LilyPond multiplier.
    """

    skip_1 = abjad.Skip("s4", multiplier=(1, 2))
    skip_2 = copy.copy(skip_1)

    assert isinstance(skip_1, abjad.Skip)
    assert isinstance(skip_2, abjad.Skip)
    assert abjad.lilypond(skip_1) == abjad.lilypond(skip_2)
    assert skip_1 is not skip_2


def test_Skip___copy___03():
    """
    Copy skip with LilyPond grob overrides and LilyPond context settings.
    """

    skip_1 = abjad.Skip((1, 4))
    abjad.override(skip_1).Staff.NoteHead.color = "#red"
    abjad.override(skip_1).Accidental.color = "#red"
    abjad.setting(skip_1).tupletFullLength = True
    skip_2 = copy.copy(skip_1)

    assert isinstance(skip_1, abjad.Skip)
    assert isinstance(skip_2, abjad.Skip)
    assert abjad.lilypond(skip_1) == abjad.lilypond(skip_2)
    assert skip_1 is not skip_2


def test_Skip___eq___01():
    skip_1 = abjad.Skip((1, 4))
    skip_2 = abjad.Skip((1, 4))
    skip_3 = abjad.Skip((1, 8))

    assert not skip_1 == skip_2
    assert not skip_1 == skip_3
    assert not skip_2 == skip_3


def test_Skip___init___01():
    """
    Initialize skip from LilyPond input string.
    """

    skip = abjad.Skip("s8.")
    assert isinstance(skip, abjad.Skip)


def test_Skip___init___02():
    """
    Initialize skip from containerize note.
    """

    chord = abjad.Chord([2, 3, 4], (1, 4))
    duration = chord.written_duration
    skip = abjad.Skip(chord)
    assert isinstance(skip, abjad.Skip)
    # check that attributes have not been removed or added.
    assert dir(chord) == dir(abjad.Chord([2, 3, 4], (1, 4)))
    assert dir(skip) == dir(abjad.Skip((1, 4)))
    assert skip._parent is None
    assert skip.written_duration == duration


def test_Skip___init___03():
    """
    Initialize skip from tupletized note.
    """

    chord = abjad.Chord([2, 3, 4], (1, 4))
    chords = abjad.mutate.copy(chord, 3)
    tuplet = abjad.Tuplet((2, 3), chords)
    duration = tuplet[0].written_duration
    skip = abjad.Skip(tuplet[0])
    assert isinstance(tuplet[0], abjad.Chord)
    assert isinstance(skip, abjad.Skip)
    assert tuplet[0]._parent is tuplet
    assert tuplet[0].written_duration == duration
    assert skip._parent is None


def test_Skip___init___04():
    """
    Initialize skip from beamed chord.
    """

    chord = abjad.Chord([2, 3, 4], (1, 4))
    chords = abjad.mutate.copy(chord, 3)
    staff = abjad.Staff(chords)
    abjad.beam(staff[:])
    skip = abjad.Skip(staff[0])
    assert isinstance(staff[0], abjad.Chord)
    assert isinstance(skip, abjad.Skip)
    assert staff[0]._parent is staff
    assert skip._parent is None


def test_Skip___init___05():
    note = abjad.Note(2, (1, 8))
    duration = note.written_duration
    skip = abjad.Skip(note)
    assert isinstance(skip, abjad.Skip)
    # check that attributes have not been removed or added.
    assert dir(note) == dir(abjad.Note("c'4"))
    assert dir(skip) == dir(abjad.Skip((1, 4)))
    assert abjad.lilypond(skip) == "s8"
    assert skip._parent is None
    assert skip.written_duration == duration


def test_Skip___init___06():
    tuplet = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    duration = tuplet[0].written_duration
    skip = abjad.Skip(tuplet[0])
    assert isinstance(tuplet[0], abjad.Note)
    assert isinstance(skip, abjad.Skip)
    assert tuplet[0]._parent is tuplet
    assert tuplet[0].written_duration == duration


def test_Skip___init___07():
    """
    Initialize skip from beamed note.
    """

    voice = abjad.Voice("c'8 c'8 c'8")
    abjad.beam(voice[:])
    skip = abjad.Skip(voice[0])
    assert isinstance(voice[0], abjad.Note)
    assert isinstance(skip, abjad.Skip)
    assert voice[0]._parent is voice


def test_Skip___init___08():
    """
    Initialize skip from unincorporaed rest.
    """

    rest = abjad.Rest((1, 8))
    duration = rest.written_duration
    skip = abjad.Skip(rest)
    assert isinstance(skip, abjad.Skip)
    # check that attributes have not been removed or added.
    assert dir(rest) == dir(abjad.Rest((1, 4)))
    assert dir(skip) == dir(abjad.Skip((1, 4)))
    assert skip._parent is None
    assert skip.written_duration == duration


def test_Skip___init___09():
    """
    Initialize skip from tupletized rest.
    """

    tuplet = abjad.Tuplet((2, 3), "r8 r8 r8")
    duration = tuplet[0].written_duration
    skip = abjad.Skip(tuplet[0])
    assert isinstance(skip, abjad.Skip)
    assert isinstance(tuplet[0], abjad.Rest)
    assert tuplet[0]._parent is tuplet
    assert tuplet[0].written_duration == duration
    assert skip._parent is None


def test_Skip___init___10():
    """
    Initialize skip from spanned rest.
    """

    voice = abjad.Voice(
        [abjad.Note(0, (1, 8)), abjad.Rest((1, 8)), abjad.Note(0, (1, 8))]
    )
    abjad.beam(voice[:])
    skip = abjad.Skip(voice[1])
    assert isinstance(skip, abjad.Skip)
    assert isinstance(voice[1], abjad.Rest)
    assert voice[1]._parent is voice
    assert skip._parent is None


def test_Skip___ne___01():
    skip_1 = abjad.Skip((1, 4))
    skip_2 = abjad.Skip((1, 4))
    skip_3 = abjad.Skip((1, 8))

    assert skip_1 != skip_2
    assert skip_1 != skip_3
    assert skip_2 != skip_3



import enum
import typing

from . import duration as _duration
from . import get as _get
from . import indicators as _indicators
from . import instruments as _instruments
from . import iterate as _iterate
from . import iterpitches as _iterpitches
from . import parentage as _parentage
from . import score as _score
from . import sequence as _sequence
from . import tag as _tag


def _aggregate_context_wrappers(argument):
    """
    Special_Voice may contain other instances of Special_Voice.
    This currently happens with OnBeatGraceContainer.
    This method aggregates all Special_Voice wrappers for checks.
    """
    context_name_to_wrappers = {}
    for context in _iterate.components(argument, _score.Context):
        if context.name not in context_name_to_wrappers:
            context_name_to_wrappers[context.name] = []
        wrappers = context._dependent_wrappers[:]
        context_name_to_wrappers[context.name].extend(wrappers)
    return context_name_to_wrappers


def check_beamed_lone_notes(argument) -> tuple[list, int]:
    r"""
    Checks beamed lone notes.

    ..  container:: example

        Beamed single notes are not wellformed:

        >>> voice = abjad.Voice("c'8 d' e' f'")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[0])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                [
                ]
                d'8
                e'8
                f'8
            }

        >>> abjad.wf.check_beamed_lone_notes(voice)
        ([Note("c'8")], 4)

    The examples above feature Abjad voice containers because beams are
    voice-persistent.
    """
    violators, total = [], 0
    for leaf in _iterate.leaves(argument):
        total += 1
        if _get.has_indicator(leaf, _indicators.StartBeam):
            if _get.has_indicator(leaf, _indicators.StopBeam):
                violators.append(leaf)
    return violators, total


def check_beamed_long_notes(argument) -> tuple[list, int]:
    r"""
    Checks beamed long notes.

    ..  container:: example

        Beamed quarter notes are not wellformed:

        >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                [
                d'4
                ]
                e'4
                f'4
            }

        >>> abjad.wf.check_beamed_long_notes(voice)
        ([Note("c'4"), Note("d'4")], 4)

        Beamed eighth notes are wellformed:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                [
                d'8
                ]
                e'8
                f'8
            }

        >>> abjad.wf.check_beamed_long_notes(voice)
        ([], 4)

    The examples above feature Abjad voice containers because beams are
    voice-persistent.
    """
    violators, total = [], 0
    for leaf in _iterate.leaves(argument):
        total += 1
        if leaf.written_duration < _duration.Duration((1, 4)):
            continue
        start_wrapper = _get.effective(leaf, _indicators.StartBeam, unwrap=False)
        if start_wrapper is None:
            continue
        stop_wrapper = _get.effective(leaf, _indicators.StopBeam, unwrap=False)
        if stop_wrapper is None:
            violators.append(leaf)
            continue
        if stop_wrapper.leaked_start_offset < start_wrapper.leaked_start_offset:
            violators.append(leaf)
            continue
        leaf_start_offset = leaf._get_timespan().start_offset
        if stop_wrapper.leaked_start_offset == leaf_start_offset:
            violators.append(leaf)
    return violators, total


def check_duplicate_ids(argument) -> tuple[list, int]:
    """
    Checks duplicate IDs.
    """
    violators = []
    components = _iterate.components(argument)
    total_ids = [id(_) for _ in components]
    unique_ids = _sequence.remove_repeats(total_ids)
    if len(unique_ids) < len(total_ids):
        for current_id in unique_ids:
            if 1 < total_ids.count(current_id):
                violators.extend([_ for _ in components if id(_) == current_id])
    return violators, len(total_ids)


def check_empty_containers(argument) -> tuple[list, int]:
    r"""
    Checks empty containers.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> staff.append(abjad.Container())

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            c'4
            d'4
            e'4
            f'4
            {
            }
        }

        >>> violators, total = abjad.wf.check_empty_containers(staff)
        >>> violators
        [Container()]

    """
    violators, containers = [], set()
    for container in _iterate.components(argument, _score.Container):
        containers.add(container)
        if len(container) == 0:
            violators.append(container)
    return violators, len(containers)


def check_missing_parents(argument) -> tuple[list, int]:
    """
    Checks missing parents.
    """
    violators, total = [], set()
    components = _iterate.components(argument)
    for i, component in enumerate(components):
        total.add(component)
        if 0 < i:
            # BeforeGraceContainer._parent is always none;
            # so must use Parentage.parent,
            # which calls BeforeGraceContainer._get_parentage()
            if _parentage.Parentage(component).parent is None:
                violators.append(component)
    return violators, len(total)


def check_notes_on_wrong_clef(argument) -> tuple[list, int]:
    r"""
    Checks notes and chords on wrong clef.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> clef = abjad.Clef('alto')
        >>> abjad.attach(clef, staff[0])
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                c'8
                d'8
                e'8
                f'8
            }

        >>> abjad.wf.check_notes_on_wrong_clef(staff)
        ([Note("c'8"), Note("d'8"), Note("e'8"), Note("f'8")], 4)

    ..  container:: example

        All instruments allow percussion clef:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> clef = abjad.Clef('percussion')
        >>> abjad.attach(clef, staff[0])
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "percussion"
                c'8
                d'8
                e'8
                f'8
            }

        >>> abjad.wf.check_notes_on_wrong_clef(staff)
        ([], 4)

    """
    violators, total = [], set()
    for leaf in _iterate.leaves(argument):
        total.add(leaf)
        instrument = _get.effective(leaf, _instruments.Instrument)
        if instrument is None:
            continue
        effective_clef = _get.effective(leaf, _indicators.Clef)
        if effective_clef is None:
            continue
        clefs = []
        for clef in instrument.clefs:
            if isinstance(clef, str):
                clef = _indicators.Clef(clef)
            assert isinstance(clef, _indicators.Clef), repr(clef)
            clefs.append(clef)
        clefs.append(_indicators.Clef("percussion"))
        if effective_clef not in clefs:
            violators.append(leaf)
    return violators, len(total)


def check_orphaned_dependent_wrappers(argument) -> tuple[list, int]:
    r"""
    Checks orphaned dependent wrappers.

    This should normally never happen because Abjad manages dependent wrappers
    behind the scenes.

    This check exists to make sure that any new code added to Abjad doesn't
    accidentally mangle dependent-wrapper handling.

    ..  container:: example

        >>> voice = abjad.Voice("c'8 [ d' e' f'")
        >>> assert len(voice._dependent_wrappers) == 1
        >>> wrapper = voice._dependent_wrappers[0]
        >>> wrapper
        Wrapper(annotation=None, context='Voice', deactivate=False, direction=None, indicator=StartBeam(), synthetic_offset=None, tag=Tag(string=''))

        >>> wrapper.component
        Note("c'8")

        >>> abjad.wf.check_orphaned_dependent_wrappers(voice)
        ([], 1)

        >>> voice[0:1] = [abjad.Note("cs'8")]
        >>> voice._dependent_wrappers
        []

        >>> abjad.wf.check_orphaned_dependent_wrappers(voice)
        ([], 0)

        >>> voice._dependent_wrappers.append(wrapper)
        >>> assert len(voice._dependent_wrappers) == 1
        >>> assert wrapper.component not in voice

        >>> abjad.wf.check_orphaned_dependent_wrappers(voice)
        ([Wrapper(annotation=None, context='Voice', deactivate=False, direction=None, indicator=StartBeam(), synthetic_offset=None, tag=Tag(string=''))], 1)

    """
    violators, total = [], 0
    for context in _iterate.components(argument, _score.Context):
        assert isinstance(context, _score.Context)
        for wrapper in context._dependent_wrappers:
            total += 1
            parentage = _get.parentage(wrapper.component)
            if context not in parentage:
                violators.append(wrapper)
    return violators, total


def check_out_of_range_pitches(
    argument, *, allow_indicators: typing.Sequence[str | enum.Enum] = ()
) -> tuple[list, int]:
    r"""
    Checks out-of-range notes.

    ..  container:: example

        Out of range:

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        >>> abjad.wf.check_out_of_range_pitches(staff)
        ([Chord('<d fs>8')], 2)

    ..  container:: example

        Using ``allow_indicators``:

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.attach("ALLOW_OUT_OF_RANGE", staff[2])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        Does not check for indicator:

        >>> violators, total = abjad.wf.check_out_of_range_pitches(staff)
        >>> violators
        [Chord('<d fs>8')]

        Does check for indicator:

        >>> violators, total = abjad.wf.check_out_of_range_pitches(
        ...     staff, allow_indicators=["ALLOW_OUT_OF_RANGE"]
        ... )
        >>> violators
        []

    """
    violators, total = [], set()
    for leaf in _iterate.leaves(argument, pitched=True):
        total.add(leaf)
        ok = False
        for indicator in allow_indicators or ():
            if leaf._has_indicator(indicator):
                ok = True
        if ok is True:
            continue
        if "unpitched" in argument._get_indicators(str):
            continue
        instrument = _get.effective(leaf, _instruments.Instrument)
        if instrument is None:
            continue
        if not _iterpitches.sounding_pitches_are_in_range(leaf, instrument.pitch_range):
            violators.append(leaf)
    return violators, len(total)


def check_overlapping_beams(argument) -> tuple[list, int]:
    r"""
    Checks overlapping beams.

    ..  container:: example

        >>> voice = abjad.Voice("c'8 [ d' [ e' f' ]")
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'8
            [
            d'8
            [
            e'8
            f'8
            ]
        }

        >>> abjad.wf.check_overlapping_beams(voice)
        ([Note("d'8")], 3)

        >>> voice = abjad.Voice("c'8 [ d' [ e' ] f' ]")
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'8
            [
            d'8
            [
            e'8
            ]
            f'8
            ]
        }

        >>> abjad.wf.check_overlapping_beams(voice)
        ([Note("d'8")], 4)

        >>> voice = abjad.Voice("c'8 [ d' e' f' ]")
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }

        >>> abjad.wf.check_overlapping_beams(voice)
        ([], 2)

    """
    violators, total = [], 0
    context_name_to_wrappers = _aggregate_context_wrappers(argument)
    for _, wrappers in context_name_to_wrappers.items():
        wrappers.sort(key=lambda _: _get.timespan(_.component).start_offset)
        open_beam_count = 0
        for wrapper in wrappers:
            if _get.grace(wrapper.component) is True:
                continue
            total += 1
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartBeam):
                open_beam_count += 1
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopBeam):
                open_beam_count -= 1
            if open_beam_count < 0 or 1 < open_beam_count:
                violators.append(wrapper.component)
    return violators, total


def check_overlapping_text_spanners(argument) -> tuple[list, int]:
    r"""
    Checks overlapping text spanners.

    Overlapping text spanners are not wellformed:

    ..  container:: example

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.text_spanner(voice)
        >>> abjad.text_spanner(voice[1:3])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            d'4
            \startTextSpan
            e'4
            \stopTextSpan
            f'4
            \stopTextSpan
        }

        >>> abjad.wf.check_overlapping_text_spanners(voice)
        ([Note("d'4")], 2)

    ..  container:: example

        Overlapping text spanners with different IDs are wellformed:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.text_spanner(voice)
        >>> command = r'\startTextSpanOne'
        >>> start_text_span = abjad.StartTextSpan(command=command)
        >>> abjad.attach(start_text_span, voice[1])
        >>> command = r'\stopTextSpanOne'
        >>> stop_text_span = abjad.StopTextSpan(command=command)
        >>> abjad.attach(stop_text_span, voice[2])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            d'4
            \startTextSpanOne
            e'4
            \stopTextSpanOne
            f'4
            \stopTextSpan
        }

        >>> abjad.wf.check_overlapping_text_spanners(voice)
        ([], 2)

    ..  container:: example

        Enchained text spanners do not overlap (and are wellformed):

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.text_spanner(voice[:3])
        >>> abjad.text_spanner(voice[-2:])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            d'4
            e'4
            \stopTextSpan
            \startTextSpan
            f'4
            \stopTextSpan
        }

        >>> abjad.wf.check_overlapping_text_spanners(voice)
        ([], 2)

    ..  container:: example

        REGRESSION. Matching start- and stop-text-spans on a single leaf do
        not overlap (and are wellformed) iff stop-text-span leaks to the
        right:

        >>> voice = abjad.Voice("c'2 d'2 e'2 f'2")
        >>> abjad.attach(abjad.StartTextSpan(), voice[0])
        >>> stop_text_span = abjad.StopTextSpan(leak=True)
        >>> abjad.attach(stop_text_span, voice[0])
        >>> abjad.attach(abjad.StartTextSpan(), voice[2])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[3])
        >>> abjad.show(voice) # doctest: +SKIP

        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'2
            \startTextSpan
            <> \stopTextSpan
            d'2
            e'2
            \startTextSpan
            f'2
            \stopTextSpan
        }

        >>> abjad.wf.check_overlapping_text_spanners(voice)
        ([], 2)

    """
    violators, total = [], 0

    def key(wrapper):
        if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
            priority = 1
        else:
            priority = 0
        return (wrapper.leaked_start_offset, priority)

    context_name_to_wrappers = _aggregate_context_wrappers(argument)
    for _, wrappers in context_name_to_wrappers.items():
        wrappers.sort(key=key)
        open_spanners: dict = {}
        for wrapper in wrappers:
            if wrapper.deactivate is True:
                continue
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
                total += 1
                command = wrapper.unbundle_indicator().command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                if open_spanners[command]:
                    violators.append(wrapper.component)
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopTextSpan):
                command = wrapper.unbundle_indicator().command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command in open_spanners and open_spanners[command]:
                    open_spanners[command].pop()
    return violators, total


def check_unmatched_stop_text_spans(argument) -> tuple[list, int]:
    r"""
    Checks unmatched stop text spans.

    ..  container:: example

        Unmatched stop text span is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            c'4
            c'4
            c'4
            \stopTextSpan
        }

        >>> abjad.wf.check_unmatched_stop_text_spans(voice)
        ([Note("c'4")], 0)

        Matched stop text span is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \startTextSpan
                c'4
                c'4
                c'4
                \stopTextSpan
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    context_name_to_wrappers = _aggregate_context_wrappers(argument)
    for _, wrappers in context_name_to_wrappers.items():
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        open_spanners: dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
                total += 1
                command = wrapper.unbundle_indicator().command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopTextSpan):
                command = wrapper.unbundle_indicator().command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command not in open_spanners or not open_spanners[command]:
                    violators.append(wrapper.component)
                else:
                    open_spanners[command].pop()
    return violators, total


def check_unterminated_hairpins(argument) -> tuple[list, int]:
    r"""
    Checks unterminated hairpins.

    ..  container:: example

        Unterminated crescendo is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \<
            c'4
            c'4
            c'4
        }

        >>> abjad.wf.check_unterminated_hairpins(voice)
        ([Note("c'4")], 1)

        Even with start dynamic:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> dynamic = abjad.Dynamic('f')
        >>> abjad.attach(dynamic, voice[0])
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \f
            \<
            c'4
            c'4
            c'4
        }

        >>> abjad.wf.check_unterminated_hairpins(voice)
        ([Note("c'4")], 1)

        Terminated crescendo is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> dynamic = abjad.Dynamic('f')
        >>> abjad.attach(dynamic, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \<
                c'4
                c'4
                c'4
                \f
            }

        >>> abjad.wf.wellformed(voice)
        True

    ..  container:: example

        Bang-terminated crescendo is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> stop_hairpin = abjad.StopHairpin()
        >>> abjad.attach(stop_hairpin, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \<
                c'4
                c'4
                c'4
                \!
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    context_name_to_wrappers = _aggregate_context_wrappers(argument)
    for _, wrappers in context_name_to_wrappers.items():
        last_dynamic = None
        last_tag = _tag.Tag()
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        for wrapper in wrappers:
            parameter = getattr(wrapper.unbundle_indicator(), "parameter", None)
            if parameter == "DYNAMIC" or isinstance(
                wrapper.unbundle_indicator(), _indicators.StopHairpin
            ):
                last_dynamic = wrapper.unbundle_indicator()
                last_tag = wrapper.tag
                if isinstance(wrapper.unbundle_indicator(), _indicators.StartHairpin):
                    total += 1
        if (
            isinstance(last_dynamic, _indicators.StartHairpin)
            and _tag.Tag("RIGHT_BROKEN").string not in last_tag.string
        ):
            violators.append(wrapper.component)
    return violators, total


def check_unterminated_text_spanners(argument) -> tuple[list, int]:
    r"""
    Checks unterminated text spanners.

    ..  container:: example

        Unterminated text spanner is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            c'4
        }

        >>> abjad.wf.check_unterminated_text_spanners(voice)
        ([Note("c'4")], 1)

        Terminated text span is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \startTextSpan
                c'4
                c'4
                c'4
                \stopTextSpan
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    context_name_to_wrappers = _aggregate_context_wrappers(argument)
    for _, wrappers in context_name_to_wrappers.items():
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        open_spanners: dict = {}
        for wrapper in wrappers:
            if wrapper.deactivate is True:
                continue
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
                total += 1
                command = wrapper.unbundle_indicator().command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopTextSpan):
                command = wrapper.unbundle_indicator().command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command not in open_spanners or not open_spanners[command]:
                    # unmatched stop text span
                    pass
                else:
                    open_spanners[command].pop()
        for command, list_ in open_spanners.items():
            for component in list_:
                violators.append(component)
    return violators, total


_globals = globals()


def _call_functions(
    component,
    check_beamed_lone_notes: bool = True,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_orphaned_dependent_wrappers: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_beams: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
):
    triples = []
    if check_beamed_lone_notes:
        name = "check_beamed_lone_notes"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_beamed_long_notes:
        name = "check_beamed_long_notes"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_duplicate_ids:
        name = "check_duplicate_ids"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_empty_containers:
        name = "check_empty_containers"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_missing_parents:
        name = "check_missing_parents"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_notes_on_wrong_clef:
        name = "check_notes_on_wrong_clef"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_orphaned_dependent_wrappers:
        name = "check_orphaned_dependent_wrappers"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_out_of_range_pitches:
        name = "check_out_of_range_pitches"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_overlapping_beams:
        name = "check_overlapping_beams"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_overlapping_text_spanners:
        name = "check_overlapping_text_spanners"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unmatched_stop_text_spans:
        name = "check_unmatched_stop_text_spans"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unterminated_hairpins:
        name = "check_unterminated_hairpins"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unterminated_text_spanners:
        name = "check_unterminated_text_spanners"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    return triples


def tabulate_wellformedness(
    component,
    check_beamed_lone_notes: bool = True,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_orphaned_dependent_wrappers: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_beams: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
) -> tuple[int, str]:
    """
    Tabulates wellformedness.
    """
    triples = _call_functions(
        component,
        check_beamed_lone_notes=check_beamed_lone_notes,
        check_beamed_long_notes=check_beamed_long_notes,
        check_duplicate_ids=check_duplicate_ids,
        check_empty_containers=check_empty_containers,
        check_missing_parents=check_missing_parents,
        check_notes_on_wrong_clef=check_notes_on_wrong_clef,
        check_orphaned_dependent_wrappers=check_orphaned_dependent_wrappers,
        check_out_of_range_pitches=check_out_of_range_pitches,
        check_overlapping_beams=check_overlapping_beams,
        check_overlapping_text_spanners=check_overlapping_text_spanners,
        check_unmatched_stop_text_spans=check_unmatched_stop_text_spans,
        check_unterminated_hairpins=check_unterminated_hairpins,
        check_unterminated_text_spanners=check_unterminated_text_spanners,
    )
    strings = []
    total_violators = 0
    for violators, total, name in triples:
        violator_count = len(violators)
        name = name.replace("check_", "")
        name = name.replace("_", " ")
        string = f"{violator_count} /    {total} {name}"
        strings.append(string)
        total_violators += violator_count
    return total_violators, "\n".join(strings)


def wellformed(
    component,
    check_beamed_lone_notes: bool = True,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_orphaned_dependent_wrappers: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_beams: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
) -> bool:
    """
    Is true when ``component`` is wellformed.
    """
    triples = _call_functions(
        component,
        check_beamed_lone_notes=check_beamed_lone_notes,
        check_beamed_long_notes=check_beamed_long_notes,
        check_duplicate_ids=check_duplicate_ids,
        check_empty_containers=check_empty_containers,
        check_missing_parents=check_missing_parents,
        check_notes_on_wrong_clef=check_notes_on_wrong_clef,
        check_orphaned_dependent_wrappers=check_orphaned_dependent_wrappers,
        check_out_of_range_pitches=check_out_of_range_pitches,
        check_overlapping_beams=check_overlapping_beams,
        check_overlapping_text_spanners=check_overlapping_text_spanners,
        check_unmatched_stop_text_spans=check_unmatched_stop_text_spans,
        check_unterminated_hairpins=check_unterminated_hairpins,
        check_unterminated_text_spanners=check_unterminated_text_spanners,
    )
    for violators, total, name in triples:
        if violators:
            return False
    return True
